#define MEASURE_QUEUE_FULL_WRITEx
// TODO: Add radial select mode (drive always selected)
// This code is for emulating a MFM disk. It provides the
// following functions
// CMD_START: Start emulation
// CMD_EXIT: Exit
//
// Commands are read from PRU0_CMD and we then write to the same
// location the status of the command.
// If the command needs a single word of data or returns one it is in
// PRU0_CMD_DATA
//
// Time is in 200 MHz clocks
//
// Head and select lines are on external GPIO0 module. We set up an interrupt
// so we can tell when they change without polling slow read of the GPIO0 pin
// register
//
// This PRU handles the Pulse Width Modulation (PWM) generation of 
// read data to the disk controller using PWM data words from PRU 1. It also
// handles capturing the controller write data as delta time between pulses and
// sends as delta transition time words to PRU 1. 
// This PRU also handles the head and drive select lines. Head select is 
// handled by stopping the MFM output data, and telling PRU 1 to start 
// generating data from the new head. Select is similar, when we are 
// selected we tell PRU 1 to start generating data and we turn off all 
// control signals when select goes away.
// 
// We also handle seeks. We count the number of step pulses and update
// the cylinder number. After the steps are done we interrupt the ARM so
// it can write out the current buffer if it's dirty and fetch the new
// cylinder. It sends START command after the data is updated.
//
// Disk rotation is tracked using the cycle counter. Due to processing
// delays when starting reads etc. the actual bit time will lag somewhat
// from the cycle counter. The offset is kept in TIME_OFFSET and used
// to correct the value read from the cycle counter. Our time between
// indexes will vary slightly due to the overhead. When head is switched
// or drive selected we start bitstream generation at pretty close to where 
// a real drive would start. The code may actually switch in negative time 
// due to the approximations. Slightly early shouldn't cause problems.
// Slightly late can require an extra rotation.  The write data is inserted 
// within a couple bits of the "proper" position using bit counts, not the
// cycle counter. The cycle counter is used so we can track rotation while
// not selected when we aren't generating bits.
// TODO: Verify accuracy of read/write
//
// Synchronization between the two PRU's is via XFER bank 10.
// r0 TRACK_TIME 
//    Transfers from PRU 0 the current track time 1 which PRU 1
//    uses to determine the correct memory location/bit to start converting
//    the track bits into PWM words.
// r8.w0 PRU1_BUF_OFFSET
//    Transfers from PRU 1 the offset in the shared memory of the word it
//    has last written or read. Set to 0 at the start of a track. Bit 16 set
//    indicates writing done for queue fill checks. (PRU1_BUF_STATE)
// r8.b3 PRU1_STATE
//    Transfers from PRU 1 the current operation state machine state. Used
//    to synchronize operation of the two PRU's.
// r9.w0 PRU0_BUF_OFFSET
//    Transfers from PRU 0 the offset in the shared memory of the word it
//    has last written or read. Set to 0 at the start of a track. Bit 16 set
//    indicates writing done for queue fill checks. (PRU0_BUF_STATE)
// r9.b3 PRU0_STATE
//    Transfers from PRU 0 the current operation state machine state. Used
//    to synchronize operation of the two PRU's.
// r18 TRACK_BIT
//    Transfers from PRU 1 the bit count of the first PWM word written to
//    the shared memory. Transfers from PRU 0 at start of write the bit time
//    write went active to determine where PRU 1 updates track bits.
//
// Summary of state/operation transition. See code for full details of
// operation
//
// 0: Initialize PRU0_STATE(STATE_IDLE)
// 0: Wait for go command
// 0: 0track_loop:
// 0: Write TRACK_TIME, PRU0_BUF_OFFSET, PRU0_STATE(STATE_READ)
// 1: 1track_loop:
// 1: Write PRU1_STATE(STATE_READ_DONE)
// 1: Wait PRU0_STATE(STATE_READ)
// 0: Wait PRU1_STATE(STATE_READ_FILLED)
// 1: Read TRACK_TIME, write TRACK_BIT, PRU1_STATE(STATE_READ)
// 1: 1PWM_word_loop:
// 1: Convert track word and write to shared memory and update PRU1_BUF_OFFSET
// 1: If buffer full send PRU1_STATE(STATE_READ_FILLED)
// 0: 0PWM_word_loop:
// 0: Read PWM words, update PRU0_BUF_OFFSET and send to PWM hardware
// 0: If write_gate goes active goto 0write
// 0: If seek or head change goto 0state_change
// 1: If PRU0_STATE != STATE_READ goto 1handle_state_change
// 1: If not end of data goto 1PWM_word_loop:
// 1: Write PRU word of zero (setting STATE_READ_FILLED)
// 0: If PWM word 0 not zero goto 0PWM_word_loop
// 0: Write PRU0_STATE(STATE_READ_DONE)
// 1: Wait PRU0_STATE(STATE_READ_DONE)
// 1: Write PRU1_STATE(STATE_READ_DONE)
// 1: goto 1track_loop:
// 0: Wait PRU1_STATE(STATE_READ_DONE)
// 0: Goto 0track_loop
//
// 0: 0write:
// 0: Write TRACK_BIT, PRU0_STATE(STATE_WRITE_WAIT)
// 1: 1handle_state_change:
// 1: If PRU0_STATE(STATE_READ_DONE) goto 1restart_read
// 1: IF PRU0_STATE not STATE_WRITE_WAIT halt
// 1: Read TRACK_BIT, set PRU1_STATE(STATE_WRITE_WAIT)
// 0: Wait PRU1_STATE(STATE_WRITE_WAIT)
// 0: 0write_word_loop:
// 0: Read ecap DELTA time, write to shared memory & update PRU0_BUF_OFFSET
// 0: If write not inactive goto 0write_word_loop:
// 0: goto 0write_done
// 1: 1write_word_loop:
// 1: If word available read it and update PRU1_BUF_OFFSET
// 1: If word zero goto 1write_done:
// 1: Decode and update track bits
// 1: goto 1write_word_loop:
// 0: 0write_done:
// 0: write zero delta word
// 0: wait for PRU1_STATE(STATE_WRITE_DONE)
// 0: goto 0track_loop
// 1: 1write_done:
// 1: Write PRU1_STATE(STATE_WRITE_DONE)
// 1: Wait PRU0_STATE(STATE_READ)
// 1: Goto special start track from end location of write
//
// 0: 0state_change: (handling seek/head change not show)
// 0: Write PRU0_STATE(STATE_READ_DONE)
// 0: Wait PRU1_STATE(STATE_READ_DONE)
// 0: goto 0track_loop
// 1: 1restart_read:
// 1: Write PRU1_STATE(STATE_READ_DONE)
// 1: Wait PRU0_STATE(STATE_READ_DONE)
// 1: goto 1track_loop


// Copyright 2014 David Gesswein.
// This file is part of MFM disk utilities.
//
// MFM disk utilities is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// MFM disk utilities is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MFM disk utilities.  If not, see <http://www.gnu.org/licenses/>.
// Code will need to change if this is changed

.setcallreg r29.w0

.origin 0
.entrypoint START

#include "prucode.hp"

#include "inc/cmd.h"

// These are registers that are used globally. Check registers defined in 
// cmd.h if looking for free registers.

   // PWM word from PRU 1 in various forms
#define PWM_WORD     r5
   // Cycle counter time to turn off index pulse
#define END_INDEX_TIME r6
  
   // Offset between cycle counter and simulated disk rotation
#define TIME_OFFSET  r7
   // Register value 0
#define RZERO        r19
   // Current cylinder
#define CYL          r20
   // Address and size of DDR memory region. Both not currently used
#define DDR_SIZE     r22 
#define DDR_ADDR     r23
// r25 and r26 used by subroutines

   // Minimum space left in queue from/to PRU 1. Only set if 
   // MEASURE_QUEUE_FULL read/write defined
#define MIN_QUEUE_LEFT r27
   // Address of cycle counter
#define CYCLE_CNTR   r28

   // 16.66... milliseconds (60 RPM)
#define ROTATION_TIME 3333333   
   // 200 microseconds, index pulse width
#define INDEX_WIDTH 40000
   // 500 nanoseconds, Settle time for head/select lines after change
#define SETTLE_TIME 100
   // 300 microsecond max interval between step pulses for buffered seek
#define SEEK_FINISH_TIME 60000

START:
   MOV      RZERO, 0
      // Enable OCP master port
   LBCO     r0, CONST_PRUCFG, 4, 4
      // Clear SYSCFG[STANDBY_INIT] to enable OCP master port
   CLR      r0, r0, 4         
   SBCO     r0, CONST_PRUCFG, 4, 4

      // Configure the programmable pointer register for PRU0 by setting 
      // c28_pointer[15:0]field to 0x0100.  This will make C28 point 
      // to 0x00010000 (PRU shared RAM).
   MOV      r0, 0x00000100
   MOV      r1, PRU0_CONTROL | CTPPR_0
   SBBO     r0, r1, 0, 4

      // Configure the programmable pointer register for PRU0 by 
      // setting c31_pointer[15:0] field to 0x0010.  This will make C31 point
      // to 0x80001000 (DDR memory).
   MOV      r0, 0x00100000
   MOV      r1, PRU0_CONTROL | CTPPR_1
   SBBO     r0, r1, 0, 4

      // Global enable of all host interrupts
   LDI       r0, 1
   SBCO      r0, CONST_PRUSSINTC, GER_OFFSET, 2

      // Enable host interrupt 0
   MOV       r0, HOST_NUM
   SBCO      r0, CONST_PRUSSINTC, HIESR_OFFSET, 4

      // Map channel 0 to host 0
   MOV       r1, (INTC_HOSTMAP_REGS_OFFSET + HOST_NUM)
   MOV       r0, CHN_NUM
   SBCO      r0, CONST_PRUSSINTC, r1, 1

      // Map GPIO 0 interrupt to channel 0
   MOV       r1, (INTC_CHNMAP_REGS_OFFSET + GPIO0_EVT)
   MOV       r0, CHN_NUM
   SBCO      r0, CONST_PRUSSINTC, r1, 1

      // Set to positive polarity
   MOV       r1, (INTC_REGS_BASE + INTC_SYS_INT_REGS_OFFSET)
   MOV       r0, 0xFFFFFFFF
   SBBO      r0, r1, 0, 4
   SBBO      r0, r1, 4, 4

      // Set to positive edge trigger
   SBBO      RZERO, r1, 0x80, 4
   SBBO      RZERO, r1, 0x84, 4

      // Make sure the GPIO 0 system interrupt is cleared
   MOV       r0, GPIO0_EVT
   SBCO      r0, CONST_PRUSSINTC, SICR_OFFSET, 4

      // Enable GPIO 0 system interrupt
   SBCO      r0, CONST_PRUSSINTC, EISR_OFFSET, 4

      // Enable cycle counter. We use this for timing the track/index
      // 4 cycles to read
   MOV      CYCLE_CNTR, PRU0_CONTROL | CYCLE
   MOV      r3, PRU0_CONTROL | CONTROL    
   LBBO     r2, r3, 0, 4            // Read reg
   SET      r2, r2, 3
   SBBO     r2, r3, 0, 4            // Write reg
   SBBO     RZERO, CYCLE_CNTR, 0, 4          // Clear timer

      // Enable Industrial ethernet counter. We use this for delays etc
      // It is slow (12 cycles) to read. TODO Might be able to use compare regs
   MOV      r0, 0x011 // Enable and set increment to 1
   MOV      r1, 1     // Clear overflow
   MOV      r2, 0     // Compensation disabled
   MOV      r3, 0     // Clear count
   SBCO     r0, CONST_IEP, 0, 16

      // Get DDR address and size.
   LBCO     DDR_ADDR, CONST_PRURAM, PRU_DDR_ADDR, 4
   LBCO     DDR_SIZE, CONST_PRURAM, PRU_DDR_SIZE, 4


      // Clear various variable
   MOV      CYL, 0
   MOV      TIME_OFFSET, 0
   SBCO     RZERO, CONST_PRURAM, PRU0_HEAD_SELECT_GLITCH_VALUE, 4
   SBCO     RZERO, CONST_PRURAM, PRU0_HEAD_SELECT_GLITCH_COUNT, 4
   SBCO     RZERO, CONST_PRURAM, PRU0_CUR_CYL, 4
   SBCO     RZERO, CONST_PRURAM, PRU0_CUR_HEAD, 4
   SBCO     RZERO, CONST_PRURAM, PRU0_CUR_SELECT_HEAD, 4
   SBCO     RZERO, CONST_PRURAM, PRU0_RQUEUE_UNDERRUN, 4
   SBCO     RZERO, CONST_PRURAM, PRU0_WQUEUE_OVERRUN, 4
   SBCO     RZERO, CONST_PRURAM, PRU0_SEEK_TIME, 4
   SBCO     RZERO, CONST_PRURAM, PRU_TEST0, 4
   SBCO     RZERO, CONST_PRURAM, PRU_TEST1, 4
   SBCO     RZERO, CONST_PRURAM, PRU_TEST2, 4
   SBCO     RZERO, CONST_PRURAM, PRU_TEST3, 4
   SBCO     RZERO, CONST_PRURAM, PRU_TEST4, 4
   SBCO     RZERO, CONST_PRURAM, PRU0_ECAP_OVERRUN, 4

      // Set the initial state of the various control lines. Inactive is high
      // for drive signals.
   SET      r30, R30_SEEK_COMPLETE_BIT
   SET      r30, R30_READY_BIT    
   SET      r30, R30_DRIVE_SEL  
   SET      r30, R30_WRITE_FAULT_BIT
   SET      r30, R30_INDEX_BIT
   MOV      r0, 1 << GPIO0_TRACK_0
   MOV      r1, GPIO0 | GPIO_SETDATAOUT
   SBBO     r0, r1, 0, 4
      // Turn off receiving MFM write data in so we can send data out
   CLR      r30, R30_MFM_IN_ENABLE       

   MOV      PRU0_BUF_STATE, 0
   MOV      PRU0_STATE, STATE_IDLE
   XOUT     10, PRU0_BUF_STATE, 4

#ifdef MEASURE_QUEUE_FULL_READ
   MOV      MIN_QUEUE_LEFT, (SHARED_PWM_READ_MASK+1)
   SBCO     MIN_QUEUE_LEFT, CONST_PRURAM, PRU_TEST3, 4
#endif
#ifdef MEASURE_QUEUE_FULL_WRITE
   MOV      MIN_QUEUE_LEFT, (SHARED_DELTAS_WRITE_MASK+1)
   SBCO     MIN_QUEUE_LEFT, CONST_PRURAM, PRU_TEST3, 4
#endif

   MOV      r0, CMD_STATUS_OK
   SBCO     r0, CONST_PRURAM, PRU0_CMD, 4
   SBCO     RZERO, CONST_PRURAM, PRU0_STATUS, 4

// Wait for initial go from the ARM before we start acting like a drive
// We need various data set up by it before we can start
wait_initial_cmd:
      // Keep our time cleared. Cycle counter stops if it overflows
   SBBO     RZERO, CYCLE_CNTR, 0, 4
   SBCO     RZERO, CONST_IEP, IEP_COUNT, 4
   LBCO     r1, CONST_PRURAM, PRU0_CMD, 4 
   QBEQ     EXIT, r1, CMD_EXIT
   QBEQ     mfm_setup, r1, CMD_START
   JMP      wait_initial_cmd

mfm_setup:
      // Enable GPIO lines to interrupt on rising and falling edge.
      // We want to know when they change.
      // Get which select line is ours and set only it to interrupt.
      // Set all the head select lines to interrupt
      // We OR our bits with current register contents to avoid interfering
      // with other users. Minor race conditions between read and write.
   LBCO      r0, CONST_PRURAM, PRU0_DRIVE_SELECT, 4
   MOV       r2, GPIO_DRIVE_SELECT_HEAD_LINES
   LBCO      r1, CONST_PRURAM, PRU0_NUM_HEAD, 4
   // If greater than 8 use all head select. Otherwise ignore MSB which
   // is reduced write current on earlier drives
   QBLE      headok, r1, 8            
   CLR       r2, GPIO_HEAD3
headok:
   OR        r0, r0, r2
   MOV       r1, GPIO0 | GPIO_RISINGDETECT
   LBBO      r2, r1, 0, 4
      // Combine with what's already in the register
   OR        r2, r0, r2
   SBBO      r2, r1, 0, 4
   MOV       r1, GPIO0 | GPIO_FALLINGDETECT
   LBBO      r2, r1, 0, 4
   OR        r2, r0, r2
   SBBO      r2, r1, 0, 4
      // AND CLEAR INTERRUPT 0
   MOV       r1, GPIO0 | GPIO_IRQSTATUS_0
   SBBO      r0, r1, 0, 4
      // AND ENABLE INTERRUPT 0
   MOV       r1, GPIO0 | GPIO_IRQSTATUS_SET_0
   SBBO      r0, r1, 0, 4


// Now start emulating a drive
      // Ack command
   MOV      r0, CMD_STATUS_OK
   SBCO     r0, CONST_PRURAM, PRU0_CMD, 4
   MOV      END_INDEX_TIME, INDEX_WIDTH
      // Make sure we can't trigger glitch logic in select_head
   MOV      r1, 0x12345678
   SBCO     r1, CONST_PRURAM, PRU0_CUR_SELECT_HEAD, 4
      // Not waiting a command
   SBCO     RZERO, CONST_PRURAM, PRU0_WAITING_CMD, 4
      // Make sure we have the current head select etc.
   JMP      select_head

wait_cmd:
      // Let select_head know we are waiting for a command
   MOV      r0, 1
   SBCO     r0, CONST_PRURAM, PRU0_WAITING_CMD, 4
      // If needed handle select or head line change interrupt?
   QBBS     select_head, r31, 30        
      // Check for time wrap and if index pulse needs updating
   CALL     check_rotation             
   CALL     set_index
      // Get command and branch to correct routine. Exit is checked
      // in some other places in code
   LBCO     r1, CONST_PRURAM, PRU0_CMD, 4 
   QBEQ     EXIT, r1, CMD_EXIT
   QBEQ     send_mfm, r1, CMD_START
   JMP      wait_cmd

send_mfm:
   SBCO     RZERO, CONST_PRURAM, PRU0_WAITING_CMD, 4
      // This measures time from when we signaled the ARM we needed a different
      // cylinder to when we got it. For performance monitoring.
   LBCO     r0, CONST_IEP, IEP_COUNT, 4
   SBCO     r0, CONST_PRURAM, PRU0_SEEK_TIME, 4
   MOV      r0, CMD_STATUS_OK
   SBCO     r0, CONST_PRURAM, PRU0_CMD, 4

next_mfm:
      // Turn on signals (active low). These are turned off if
      // drive not selected which then goes here when selected again
   CALL     set_track0
   CLR      r30, R30_READY_BIT    
   CLR      r30, R30_SEEK_COMPLETE_BIT 
      // Switch to PWM mode
   SBCO     RZERO, CONST_ECAP, ECCTL1, 4
   MOV      r0, 40                  // Pulse period
   SBCO     r0, CONST_ECAP, CAP1, 4
   SBCO     r0, CONST_ECAP, CAP3, 4
   SBCO     RZERO, CONST_ECAP, CAP2, 4
   SBCO     RZERO, CONST_ECAP, CAP4, 4
   SBCO     RZERO, CONST_ECAP, TSCTR, 4
      // Set APWM mode output high
   MOV      r0, 0x02100000          // low 16 is ECCTL1, upper ECCTL2
   SBCO     r0, CONST_ECAP, ECCTL1, 4

sendtrack:   
   LBBO     TRACK_TIME, CYCLE_CNTR, 0, 4   // get time
   ADD      TRACK_TIME, TRACK_TIME, TIME_OFFSET    // Adjust for offset
      // Fudge to start at beginning of track. Cover delay from restart to here
   SUB      TRACK_TIME, TRACK_TIME, 60   
   QBBC     outit, TRACK_TIME, 31
   MOV      TRACK_TIME, 0                  // Set to zero if negative
outit:
   XOUT     10, TRACK_TIME, 4
   MOV      PRU0_STATE, STATE_READ
   MOV      PRU0_BUF_OFFSET, 0
   XOUT     10, PRU0_BUF_STATE, 4
waitfill:
      // Wait for queue filled
   XIN      10, PRU1_BUF_STATE, 4              
   QBNE     waitfill, PRU1_STATE, STATE_READ_FILLED
   XIN      10, TRACK_BIT, 4
      // Calculate offset between timer and bit count for accurate track of our
      // simulated rotation which drifts slightly from real time. We can only 
      // reset the counter to zero so we can't fix counter time. IEP may be 
      // writable but is slow (12 vs 4) cycles to read
   LSL      r0, TRACK_BIT, 4       // multiply by 16
   LSL      r1, TRACK_BIT, 2       // multiply by 4
   ADD      r0, r0, r1             // Add to get multiply by 20 (clock/bit)
   LBBO     r1, CYCLE_CNTR, 0, 4   // Get current time
   SUB      TIME_OFFSET, r0, r1    // Offset between current time and bit time
   CALL     set_index

      // Send read MFM data to the controller.
      // This reads the PWM words representing the bits from PRU 1 and 
      // writes them to the PWM controller. Bits 31-28 are the number
      // of MFM bit times the word generates. Bits 27-24 should be ignored.
      // Needs to be less than 40. currently around 27 cycles
read:
   LBCO     PWM_WORD, CONST_PRUSHAREDRAM, PRU0_BUF_OFFSET, 4 // 3 cycles
   QBEQ     done, PWM_WORD, 0              // Time 0 marks end of data
   LSR      r0, PWM_WORD, 28
   ADD      TRACK_BIT, TRACK_BIT, r0       // Update bit count
   MOV      PWM_WORD.b3, 0                 // Clear count
   ADD      PRU0_BUF_OFFSET, PRU0_BUF_OFFSET, 4 
   AND      PRU0_BUF_OFFSET, PRU0_BUF_OFFSET, SHARED_PWM_READ_MASK   // Wrap if needed 
      // Send our read offset into shared memory to PRU 1 and get its write offset
   XOUT     10, PRU0_BUF_STATE, 4
   XIN      10, PRU1_BUF_STATE, 4
      // Either calculate how much space is in queue or simple did it wrap check
#ifdef MEASURE_QUEUE_FULL_READ
      // MSB set indicates writer done so don't check
   QBBS     chke, PRU1_BUF_STATE, 16                  
      // Calculate words in queue handling queue wrap
   SUB      r0, PRU1_BUF_OFFSET, PRU0_BUF_OFFSET
   QBBC     notneg, r0, 31
   ADD      r0, r0, (SHARED_PWM_READ_MASK+1)
notneg:
   QBGE     chke, MIN_QUEUE_LEFT, r0
   MOV      MIN_QUEUE_LEFT, r0
   SBCO     MIN_QUEUE_LEFT, CONST_PRURAM, PRU_TEST3, 4
   SBCO     PRU0_BUF_STATE, CONST_PRURAM, PRU_TEST1, 4
   SBCO     PRU1_BUF_STATE, CONST_PRURAM, PRU_TEST2, 4
#else
   QBNE     notfullr, PRU0_BUF_OFFSET, PRU1_BUF_OFFSET
   LBCO     r0, CONST_PRURAM, PRU0_RQUEUE_UNDERRUN, 4
   ADD      r0, r0, 1
   SBCO     r0, CONST_PRURAM, PRU0_RQUEUE_UNDERRUN, 4
   SBCO     PRU0_BUF_STATE, CONST_PRURAM, PRU_TEST1, 4
   SBCO     PRU1_BUF_STATE, CONST_PRURAM, PRU_TEST2, 4
notfullr:
#endif

chke:
   LBCO     r0, CONST_ECAP, ECFLG, 2    // did PWM read shadow regs? 4 cycles
   QBBS     loadit, r0, 6               // Yes, load a new one
   QBBC     write, r31, R31_WRITE_GATE  // If write gate active handle it
   QBBS     select_head, r31, 30        // Handle gpio interrupt
   QBBC     step, r31, R31_STEP_BIT     // Doing a seek?
   LBBO     r4, CYCLE_CNTR, 0, 4        // get time, 4 cycles
   QBLT     chke, END_INDEX_TIME, r4    // Did we reach end of index time
   SET      r30, R30_INDEX_BIT          // Yes, turn off index
   JMP      chke                        // Check PWM again
loadit:
   SBCO     PWM_WORD.w0, CONST_ECAP, CAP3, 2  // Update shadow period, 2 cycles
   SBCO     PWM_WORD.w2, CONST_ECAP, CAP4, 2  // Update shadow width, 2 cycles
   SBCO     r0, CONST_ECAP, ECCLR, 2          // Clear flag, 2 cycles
   JMP      read
done:
   SBCO     RZERO, CONST_ECAP, CAP2, 4  // Turn off MFM pulses
      // Tell PRU1 we are done and wait for it to indicate done
   MOV      PRU0_STATE, STATE_READ_DONE
   XOUT     10, PRU0_BUF_STATE, 4
done_lp:
   XIN      10, PRU1_BUF_STATE, 4
   QBNE     done_lp, PRU1_STATE, STATE_READ_DONE
      // We are starting a new track at the beginning to clear time offset
   MOV      TIME_OFFSET, 0
   SBBO     RZERO, CYCLE_CNTR, 0, 4     // Set time to beginning of track
   LBCO     r1, CONST_PRURAM, PRU0_CMD, 4 
   QBNE     sendtrack, r1, CMD_EXIT
   JMP      EXIT

      // Handle GPIO interrupt from head or select lines changing
select_head:
      // Wait a little bit for the lines to settle
   MOV      r1, SETTLE_TIME                 // How long to wait
   SBCO     RZERO, CONST_IEP, IEP_COUNT, 4  // Clear time
settle_lp:
   LBCO     r4, CONST_IEP, IEP_COUNT, 4     // Get time
   QBLT     settle_lp, r1, r4               // Did we reach settle time, no
      // Clear GPIO 0 interrupt before reading value so if it changes we will
      // get a new interrupt and not miss a change
   MOV      r1, GPIO0 | GPIO_IRQSTATUS_0
   LBBO     r4, r1, 0, 4
   SBBO     r4, r1, 0, 4
      // This read seems to be needed to prevent extra interrupts. May need
      // time for it to take effect before clearing int PRU intc
   LBBO     r4, r1, 0, 4               

      // Clear interrupt status flag in interrupt controller
   MOV      r0, GPIO0_EVT
   SBCO     r0, CONST_PRUSSINTC, SICR_OFFSET, 4

      // Copy current value and get new head and select
   LBCO     r1, CONST_PRURAM, PRU0_CUR_SELECT_HEAD, 4
   CALL     get_select_head
      // If they are the same report a glitch for debugging
   QBEQ     glitch, r24, r1

      // Glitch comes back here after handling it
restart:
   MOV      PRU0_STATE, STATE_READ_DONE
   XOUT     10, PRU0_BUF_STATE, 4
restart_lp:
   XIN      10, PRU1_BUF_STATE, 4
   QBNE     restart_lp, PRU1_STATE, STATE_READ_DONE
      // If we we aren't selected anymore handle it
   LBCO     r0, CONST_PRURAM, PRU0_DRIVE_SELECT, 4
   QBEQ     selected, r0, 0        // If zero always selected (radial select)
   QBBS     notsel, r24, r0
selected:
      // We are selected so turn on selected signal
   CLR      r30, R30_DRIVE_SEL  
      // If we are waiting for a command go back and wait for it else start
      // outputing MFM data
   LBCO     r0, CONST_PRURAM, PRU0_WAITING_CMD, 4
   QBEQ     next_mfm, r0, 0
   JMP      wait_cmd

      // We aren't selected, turn off signals and wait for select again
notsel:
      // Turn off all the control signals since we aren't selected
   SET      r30, R30_SEEK_COMPLETE_BIT
   SET      r30, R30_READY_BIT    
   SET      r30, R30_DRIVE_SEL  
   SET      r30, R30_INDEX_BIT
   MOV      r0, 1 << GPIO0_TRACK_0
   MOV      r1, GPIO0 | GPIO_SETDATAOUT
   SBBO     r0, r1, 0, 4
   SBCO     RZERO, CONST_ECAP, CAP2, 4  // Turn off data
waitsel:
   LBCO     r1, CONST_PRURAM, PRU0_CMD, 4 
   QBEQ     EXIT, r1, CMD_EXIT          // Handle exit if needed
      // Keep track of drive rotation while waiting
   CALL     check_rotation             
   QBBS     select_head, r31, 30        // Select or head line change interrupt?
   JMP      waitsel

glitch:
   SBCO     r4, CONST_PRURAM, PRU0_HEAD_SELECT_GLITCH_VALUE, 4
   LBCO     r1, CONST_PRURAM, PRU0_HEAD_SELECT_GLITCH_COUNT, 4
   ADD      r1, r1, 1
   SBCO     r1, CONST_PRURAM, PRU0_HEAD_SELECT_GLITCH_COUNT, 4
   JMP      restart

      // Handle seeks. We count the number of step pulses that are within
      // SEEK_FINISH_TIME and add/subtract from current cylinder based on
      // direction signal. After we have final cylinder we tell the ARM to
      // get us the cylinder data. 
step:
   SET      r30, R30_SEEK_COMPLETE_BIT        // Indicate seek in progress
   SBCO     RZERO, CONST_ECAP, CAP2, 4        // Turn off data

   QBBS     step_out, R31, R31_SEEK_DIR_BIT   // Which direction
      // Direction is in, higher cyl, Limit cylinder value to number of cyl-1
   LBCO     r0, CONST_PRURAM, PRU0_NUM_CYL, 4  
   SUB      r0, r0, 1  
   QBEQ     step_done, CYL, r0                // At limit, ignore
   ADD      CYL, CYL, 1
   JMP      step_done
step_out:
   QBEQ     step_done, CYL, 0                // At limit, ignore
   SUB      CYL, CYL, 1
step_done:
   SBCO     CYL, CONST_PRURAM, PRU0_CUR_CYL, 4
   MOV      PRU0_STATE, STATE_READ_DONE
   XOUT     10, PRU0_BUF_STATE, 4
step_done_lp:
   XIN      10, PRU1_BUF_STATE, 4
   QBNE     step_done_lp, PRU1_STATE, STATE_READ_DONE
waitstephigh:
      // Keep our time and index signal updated while waiting
   CALL     check_rotation             
   CALL     set_index
   QBBC     waitstephigh, r31, R31_STEP_BIT  
      // Update track zero signal in case it changed
   CALL     set_track0
      // Reset timer for step pulse timeout
   SBCO     RZERO, CONST_IEP, IEP_COUNT, 4
      // Wait this long for more pulse for buffered seek
   MOV      r1, SEEK_FINISH_TIME
seek_lp:
   QBBC     step, r31, R31_STEP_BIT  // Got another pulse
   CALL     check_rotation             
   CALL     set_index
   LBCO     r0, CONST_IEP, IEP_COUNT, 4
   QBLT     seek_lp, r1, r0

      // Send notification to Host to read next cylinder
   MOV       r31.b0, PRU0_ARM_INTERRUPT+16
      // Reset timer for seek time measurement
   SBCO     RZERO, CONST_IEP, IEP_COUNT, 4
   JMP      wait_cmd
   
      // Set/clear index signal based on rotation time
set_index:
   LBBO     r25, CYCLE_CNTR, 0, 4
   ADD      r25, r25, TIME_OFFSET     // Adjust for offset
   QBGT     setind, END_INDEX_TIME, r25
   CLR      r30, R30_INDEX_BIT        // Turn on index (active low)
   RET
setind:
   SET      r30, R30_INDEX_BIT        // Turn off index (active low)
   RET

      // Reset timer if we passed end of track time
check_rotation:
   MOV      r24, ROTATION_TIME
   LBBO     r25, CYCLE_CNTR, 0, 4     // Get time
   ADD      r25, r25, TIME_OFFSET     // Adjust for offset
   QBGT     clr_time, r24, r25        // If at disk rotation time reset time
   RET
clr_time:
   SBBO     RZERO, CYCLE_CNTR, 0, 4   // Clear time
   MOV      TIME_OFFSET, 0            // And offset
   RET

      // Set track 0 signal based on current cylinder
set_track0:
   MOV      r24, 1 << GPIO0_TRACK_0
   QBEQ     track0, CYL, 0
   MOV      r25, GPIO0 | GPIO_SETDATAOUT
   JMP      settrack0
track0:
   MOV      r25, GPIO0 | GPIO_CLEARDATAOUT
settrack0:
   SBBO     r24, r25, 0, 4
   RET

// Get select and head value
// Returns PRU0_CUR_SELECT_HEAD in r24
get_select_head:
   MOV      r25, GPIO0 | GPIO_DATIN
   LBBO     r24, r25, 0, 4                   // Read bits
   MOV      r25, GPIO_DRIVE_SELECT_HEAD_LINES
   AND      r24, r24, r25                    // Our bits
   LBCO     r25, CONST_PRURAM, PRU0_NUM_HEAD, 4
   // If number of heads is <= 8 set the MSB head select line.
   // which is reduced write current on earlier drives.
   QBLE     headok2, r1, 8            
   SET      r24, GPIO_HEAD3
headok2:
   SBCO     r24, CONST_PRURAM, PRU0_CUR_SELECT_HEAD, 4
   MOV      r25, GPIO_DRIVE_HEAD_LINES
   AND      r24, r24, r25                    // Our head bits
   XOR      r24, r24, r25                    // All high is drive 0 so invert
   LSR      r24, r24, GPIO_HEAD0             // Move to low bits
   LBCO     r25, CONST_PRURAM, PRU0_NUM_HEAD, 4
   SUB      r25, r25, 1                      // Convert to maximum value
   QBLE     storeh1, r25, r24
   MOV      r24, r25                         // Greater than max so set to max
storeh1:
   SBCO     r24, CONST_PRURAM, PRU0_CUR_HEAD, 4
   LBCO     r24, CONST_PRURAM, PRU0_CUR_SELECT_HEAD, 4
   RET

      // Capture write data from the controller.
      // We switch the ECAP from PWM mode to capture mode
write:
   XOUT     10, TRACK_BIT, 4        // Send current bit count.
   MOV      PRU0_STATE, STATE_WRITE_WAIT
   MOV      PRU0_BUF_OFFSET, 0
   XOUT     10, PRU0_BUF_STATE, 4 
wait_pru1_write:
      // Loop until PRU1 switches to ready to take write data
   XIN      10, PRU1_BUF_STATE, 4
   QBNE     wait_pru1_write, PRU1_STATE, STATE_WRITE_WAIT
   MOV      PRU0_STATE, STATE_WRITE
   XOUT     10, PRU0_BUF_STATE, 4         

      // Trigger cap1 on rising edges. Reset count on trigger to get delta time
      //Enable cap1-4 buffer registers. Divide by 1, stop on emulation suspend
      // continuous mode, wrap after 4, no rearm, counter free run, no sync
      // capture mode. Reset sequencer
   SET      r30, R30_MFM_IN_ENABLE          // Turn on receiving write data 
   SBCO     RZERO, CONST_ECAP, ECCTL1, 4    // Stop capture/pwm
   SBCO     RZERO, CONST_ECAP, TSCTR, 4     // Clear counter
   MOV      r0, 0x1e             
   SBCO     r0, CONST_ECAP, ECCLR, 2        // Clear capture flags
      // This will trigger on positive edges. 
   MOV      r0, 0x001e01aa       // Low 16 bits ECCTL1 high ECCTL2
   SBCO     r0, CONST_ECAP, ECCTL1, 4

caploop:
   LBCO     r0, CONST_ECAP, ECFLG, 2  // Read flags
      // If all flags set we may have lost a transition
   QBNE     cnext1, r0, 0x1e          
   SBCO     r0, CONST_PRURAM, PRU0_ECAP_OVERRUN, 4
cnext1:
   QBBS     get1, r0, 1               // Branch if cap1 has data
   CALL     checkstuff
   JMP      caploop
get1:
   LBCO     r10, CONST_ECAP, CAP1, 4  // Get count
   ADD      r10, r10, 1               // Count is 1 less than delta time
   MOV      r0, 0x2
   SBCO     r0, CONST_ECAP, ECCLR, 2  // Clear flag
   CALL     sendwrite
      // Repeat above for the other 3 capture registers
chk2:
   LBCO     r0, CONST_ECAP, ECFLG, 2 
   QBNE     cnext2, r0, 0x1e
   SBCO     r0, CONST_PRURAM, PRU0_ECAP_OVERRUN, 4
cnext2:
   QBBS     get2, r0, 2             
   CALL     checkstuff
   JMP      chk2
get2:
   LBCO     r10, CONST_ECAP, CAP2, 4
   ADD      r10, r10, 1
   MOV      r0, 0x4
   SBCO     r0, CONST_ECAP, ECCLR, 2
   CALL     sendwrite
chk3:
   LBCO     r0, CONST_ECAP, ECFLG, 2
   QBNE     cnext3, r0, 0x1e
   SBCO     r0, CONST_PRURAM, PRU0_ECAP_OVERRUN, 4
cnext3:
   QBBS     get3, r0, 3
   CALL     checkstuff
   JMP      chk3
get3:
   LBCO     r10, CONST_ECAP, CAP3, 4
   ADD      r10, r10, 1
   MOV      r0, 0x8
   SBCO     r0, CONST_ECAP, ECCLR, 2
   CALL     sendwrite
chk4:
   LBCO     r0, CONST_ECAP, ECFLG, 2
   QBNE     cnext4, r0, 0x1e
   SBCO     r0, CONST_PRURAM, PRU0_ECAP_OVERRUN, 4
cnext4:
   QBBS     get4, r0, 4
   CALL     checkstuff
   JMP      chk4
get4:
   LBCO     r10, CONST_ECAP, CAP4, 4
   ADD      r10, r10, 1
   MOV      r0, 0x10
   SBCO     r0, CONST_ECAP, ECCLR, 2
   CALL     sendwrite

   CALL     check_rotation           // These are slow so only do once per loop
   CALL     set_index
   JMP      caploop

// Send write delta times to PRU 1
// r10 is word to write in
// *** Changes r1 ***
sendwrite:
      // Store word and update pointer
   SBCO     r10, CONST_PRUSHAREDRAM, PRU0_BUF_OFFSET, 4
   ADD      PRU0_BUF_OFFSET, PRU0_BUF_OFFSET, 4
   AND      PRU0_BUF_OFFSET, PRU0_BUF_OFFSET, SHARED_DELTAS_WRITE_MASK
   XOUT     10, PRU0_BUF_STATE, 4             // Send offset to PRU 1
   XIN      10, PRU1_BUF_STATE, 4             // Get PRU 1 offset
      // See READ version for description
#ifdef MEASURE_QUEUE_FULL_WRITE
   QBBS     chke, PRU1_BUF_STATE, 16
   SUB      r24, PRU0_BUF_OFFSET, PRU1_BUF_OFFSET
   QBBC     notneg, r24, 31
   ADD      r24, r24, SHARED_DELTAS_WRITE_MASK
   ADD      r24, r24, 1
notneg:
   QBGE     chke, MIN_QUEUE_LEFT, r24
   MOV      MIN_QUEUE_LEFT, r24
   SBCO     MIN_QUEUE_LEFT, CONST_PRURAM, PRU_TEST3, 4
   SBCO     PRU0_BUF_STATE, CONST_PRURAM, PRU_TEST1, 4
   SBCO     PRU1_BUF_STATE, CONST_PRURAM, PRU_TEST2, 4
#else
   QBNE     notfullw, PRU0_BUF_OFFSET, PRU1_BUF_OFFSET
   LBCO     r24, CONST_PRURAM, PRU0_WQUEUE_OVERRUN, 4
   ADD      r24, r24, 1
   SBCO     r24, CONST_PRURAM, PRU0_WQUEUE_OVERRUN, 4
   SBCO     PRU0_BUF_STATE, CONST_PRURAM, PRU_TEST1, 4
   SBCO     PRU1_BUF_STATE, CONST_PRURAM, PRU_TEST2, 4
notfullw:
#endif
   RET

      // Check any stuff we need to at each delta. 
checkstuff:
      // Assume they will turn off write gate if they change select or head
   QBBS     write_done, r31, R31_WRITE_GATE  // Branch if write gate inactive
   RET
   
      // Done write, switch back to read
write_done:
   CLR      r30, R30_MFM_IN_ENABLE       // Turn off receiving write data 
   MOV      r10, 0                       // Send zero to indicate end of data
   CALL     sendwrite
write_done_lp:
   XIN      10, PRU1_BUF_STATE, 4
   QBNE     write_done_lp, PRU1_STATE, STATE_WRITE_DONE
   JMP      next_mfm

      // ARM has requested us to exit
EXIT:
   SET      r30, R30_SEEK_COMPLETE_BIT
   SET      r30, R30_READY_BIT    
   SET      r30, R30_DRIVE_SEL  
   SET      r30, R30_WRITE_FAULT_BIT
   SET      r30, R30_INDEX_BIT
   MOV      r0, 1 << GPIO0_TRACK_0
   MOV      r1, GPIO0 | GPIO_SETDATAOUT
   SBBO     r0, r1, 0, 4
   MOV      r1, CMD_STATUS_OK
   SBCO     r1, CONST_PRURAM, PRU0_CMD, 4 // Indicate command completed ok
   // Tell PRU 1 to exit. 
   MOV      PRU0_STATE, STATE_EXIT
   XOUT     10, PRU0_BUF_STATE, 4
   MOV      CYL, -1          // -1 tells seek task to exit on ARM
   SBCO     CYL, CONST_PRURAM, PRU0_CUR_CYL, 4
      // Tell ARM to read CYL
   MOV      r31.b0, PRU0_ARM_INTERRUPT+16
wait_arm:
   // Wait for ARM to service the previous interrupt before we
   // send one for exiting
   LBCO     CYL, CONST_PRURAM, PRU0_CUR_CYL, 4
   QBEQ     wait_arm, CYL.b3, 0xff
      // Send notification to Host for program completion
   MOV      r31.b0, PRU0_ARM_INTERRUPT+16

      // Halt the processor
   HALT

      // Restarted here if the ARM wishes to abort an operation. Any needed
      // cleanup should be here to restore to a good state. Currently none
   .ORIGIN RESTART_ADDR
   MOV      r0, CMD_STATUS_OK
   SBCO     r0, CONST_PRURAM, PRU0_CMD, 4
   JMP      wait_cmd
