// This code handles for disk reads getting the MFM bitstream from DDR 
// memory and converting it to pulse width modulation (PWM) words for PRU 0.
// For write it takes the delta time between MFM transitions, decodes it into
// clock and data bits and updates the bitstream in the DDR memory.
//
// The read and write data goes through a circular buffer in shared memory
// which allows time for the occasionally very slow access to DDR memory.
// Tests show reads can be delayed up to 5 microseconds (1000 cycles).
// Writes seem much better with delays around 200 cycles. Writes are
// posted which reduces the delay. If we ever get too far behind in read
// the controller should retry so it's not fatal. If we get behind in writes
// we will corrupt the track. Words need to be written faster than 40
// cycles (200 ns) on average.
// Clock is 200 MHz, 5 nanoseconds.

// An entire cylinder is stored in the DDR memory. If the head select
// lines are changed PRU 0 will tell us to restart sending data and we
// will calculate the correct location to send the data from. PRU 0
// will tell us the starting bit of the write data. We mark the cylinder
// dirty so the ARM can write it back to disk.
// PRU 0 handles getting the data updated if the controller selects another
// cylinder.

// See prucode0.p for information on the transfer memory registers and
// communication between the two PRU's.

// TODO: Should we mark dirty at track level to allow smaller writes?
// Flash seems to be more bounded by number of write and size less important.

// Copyright 2014 David Gesswein.
// This file is part of MFM disk utilities.
//
// MFM disk utilities is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// MFM disk utilities is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MFM disk utilities.  If not, see <http://www.gnu.org/licenses/>.

.setcallreg r29.w0

.origin 0
.entrypoint START

#include "prucode.hp"

#include "inc/cmd.h"

// These are registers that are used globally. Check registers defined in 
// cmd.h if looking for free registers.

   // Current offset into DDR memory
#define DDR_OFFSET    r10
   // Words in DDR memory left to process in track
#define WORDS_LEFT    r11
   // Count of bits in track
#define BIT_COUNT     r12
   // Current data word processing from DDR
#define WORD1         r14
   // Next data word processing
#define WORD2         r15
   // Total number of words in track
#define TRACK_WORDS   r17
   // Not currently used
#define DDR_SIZE      r22
   // Physical address of start of DDR memory
#define DDR_ADDR      r23
   // Offset from start of DDR memory of track data
#define TRACK_OFFSET  r24

START:
      // Enable OCP master port
   LBCO     r0, CONST_PRUCFG, 4, 4
      // Clear SYSCFG[STANDBY_INIT] to enable OCP master port
   CLR      r0, r0, 4         
   SBCO     r0, CONST_PRUCFG, 4, 4

      // Configure the programmable pointer register by setting c25_pointer[15:0]
      // field to 0x0020.  This will make C25 point to 0x00002000 (Other PRU RAM).
   MOV      r0, 0x00000020
   MOV      r4, PRU1_CONTROL | CTBIR
   SBBO     r0, r4, 0, 4
      // Configure the programmable pointer register for PRU0 by setting 
      // c31_pointer[15:0] field to 0x0010.  This will make C31 point to 
      // 0x80001000 (DDR memory). Configure the programmable pointer register
      // for PRU0 by setting c28_pointer[15:0] field to 0x0100.  This will
      //  make C28 point to 0x00010000 (PRU shared RAM).
   MOV      r0, 0x00000100
   MOV      r4, PRU1_CONTROL | CTPPR_0
   SBBO     r0, r4, 0, 4

      // Configure the programmable pointer register for PRU0 by setting 
      // c31_pointer[15:0] field to 0x0010.  This will make C31 point to
      //  0x80001000 (DDR memory).
   MOV      r0, 0x00100000
   MOV      r4, PRU1_CONTROL | CTPPR_1
   SBBO     r0, r4, 0, 4

      // Get address and size of DDR buffer. Currently we don't use size
   LBCO     DDR_ADDR, CONST_PRURAM, PRU_DDR_ADDR, 4
   LBCO     DDR_SIZE, CONST_PRURAM, PRU_DDR_SIZE, 4

      // Make a table which we index with the four MSB of the MFM bitstream
      // to determine the next PWM word. All the words generated must have
      // a minimum duration of 40 counts. For example bit pattern 100101
      // is sent as a pulse with period 60 counts and a pulse with period 40.
      // If we find a 11 we drop the second bit since we can't generate it.
      // It's bad MFM data so shouldn't cause problems. 01 we flag as an error
      // since we should never see it unless something has gone wrong.
      // Bits 31-28 are how many bits to remove from data
      // bit  24 is flag for illegal bit pattern found
      // Period values are one less than actual period generated.
   MOV      r4, PRU1_BIT_TABLE
   MOV      r0, 0x20000027 // 0000  0 39
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x3000003b // 0001  0 59
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x20000027 // 0010  0 39
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x20000027 // 0011  0 39
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x21000027 // 0100  0 39
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x3100003b // 0101  0 39
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x21000027 // 0110  0 39
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x21000027 // 0111  0 39
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x20140027 // 1000  0 39
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x3014003b // 1001 20 59
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x20140027 // 1010 20 39
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x20140027 // 1011 20 39
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x20140027 // 1100 20 39 We drop second bit
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x3014003b // 1101 20 39
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x20140027 // 1110 20 39
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4
   MOV      r0, 0x20140027 // 1111 20 39
   SBCO     r0, CONST_PRURAM, r4, 4
   ADD      r4, r4, 4

   MOV      r2, 0
   SBCO     r2, CONST_PRURAM, PRU1_BAD_PATTERN_COUNT, 4
   SBCO     r2, CONST_PRURAM, PRU_TEST0, 4
   SBCO     r2, CONST_PRURAM, PRU_TEST1, 4
   SBCO     r2, CONST_PRURAM, PRU_TEST2, 4
   SBCO     r2, CONST_PRURAM, PRU_TEST3, 4
   SBCO     r2, CONST_PRURAM, PRU_TEST4, 4

   SBCO     r2, CONST_PRURAM, PRU1_CYL_DIRTY, 4

   MOV      PRU1_BUF_STATE, 0

wait_read:
//SET r30, 0
   // Wait for PRU0, exit if requested. 
   MOV      PRU1_STATE, STATE_READ_DONE
   MOV      PRU1_BUF_OFFSET, 0
   XOUT     10, PRU1_BUF_STATE, 4 
wait_read_lp:
   XIN      10, PRU0_BUF_STATE, 4
   QBEQ     DONE, PRU0_STATE, STATE_EXIT
   QBNE     wait_read_lp, PRU0_STATE, STATE_READ 
   // Get track byte count and convert to words, all must be same
   LBCO     TRACK_WORDS, CONST_PRURAM, PRU1_TRACK_DATA_BYTES, 4 
   LSR      TRACK_WORDS, TRACK_WORDS, 2

   MOV      DDR_OFFSET, 0
   MOV      WORDS_LEFT, TRACK_WORDS

      // Calculate start of track by multiply head by track size in bytes
      // and add in header size to get start address of data
      // Get current head from PRU 0
   LBCO     r29, CONST_PRURAM_OTHER, PRU0_CUR_HEAD, 4
   MOV      r28, TRACK_WORDS
   LSL      r28, r28, 2           // Convert to byte count
   LBCO     r0, CONST_PRURAM, PRU1_TRACK_HEADER_BYTES, 4 
   ADD      r28, r28, r0
   MOV      r25, 0                // Set multiply only mode
   XOUT     0, r25, 1
   XIN      0, R26, 8             // Get multiply results (27,26=28*29)
   ADD      DDR_OFFSET, R26,  r0  // Point after header
   MOV      TRACK_OFFSET, DDR_OFFSET

      // Calculate bits = clock counts/20. Each bit is 20 clocks
   MOV      r29, 3276             // 1/20*2^16
   XIN      10, TRACK_TIME, 4
   MOV      r28, TRACK_TIME
   MOV      r25, 0                // Set multiply only mode
   XOUT     0, r25, 1
   XIN      0, r26, 8
   LSR      r4, r26, 16           // discard low 16 bits of low result
   LSL      r27, r27, 32-16       // get proper bits of upper result
   OR       r4, r4, r27           // and combine with lower
      // Convert to word offset, divide by 2^5, 32
   LSR      r4, r4, 5
      // If more words than in data set to last word in data
   QBLT     writeoffset, WORDS_LEFT, r4  
   MOV      r4, WORDS_LEFT        
   SUB      r4, r4, 1             // Process minimum of 1 word
writeoffset:
   LSL      r0, r4, 2             // Convert word offset to byte
   ADD      DDR_OFFSET, DDR_OFFSET, r0     // Add offset into track
   SUB      WORDS_LEFT, WORDS_LEFT, r4     // Get how many words left
restart_read:
      // Convert byte offset from start of track to bits to determine where
      // we start sending data.
   SUB      TRACK_BIT, DDR_OFFSET, TRACK_OFFSET   
   LSL      TRACK_BIT, TRACK_BIT, 3        // In bytes, convert to bits
   XOUT     10, TRACK_BIT, 4               // And tell PRU 0 start bit of data
   MOV      PRU1_BUF_OFFSET, 0             // Start filling at first location
   LBBO     WORD1, DDR_ADDR, DDR_OFFSET, 8 // Get WORD1,WORD2
   ADD      DDR_OFFSET, DDR_OFFSET, 8      // Point to next word

      // Clear this bit to prevent bad pattern errors in decoding.
   CLR      WORD1, 30            
   MOV      BIT_COUNT, 31                  // 32 bits to process, negative stop
//CLR r30, 0
// TODO This is a little slow, ~150 ns/output word 200ns is limit
bitloop:
   LSR      r0, WORD1, 28                  // Get upper 4 bits of MFM data
   LSL      r0, r0, 2                      // Make byte address of word
   ADD      r0, r0, PRU1_BIT_TABLE          
   LBCO     r0, CONST_PRURAM, r0, 4        // Lookup next PWM word
   SBCO     r0, CONST_PRUSHAREDRAM, PRU1_BUF_OFFSET, 4  // Store it
   ADD      PRU1_BUF_OFFSET, PRU1_BUF_OFFSET, 4         // Move to next location
      // Wrap if needed
   AND      PRU1_BUF_OFFSET, PRU1_BUF_OFFSET, SHARED_PWM_READ_MASK   
   XOUT     10, PRU1_BUF_STATE, 4          // Send our offset
   QBBC     noerr, r0, 24
   LBCO     r2, CONST_PRURAM, PRU1_BAD_PATTERN_COUNT, 4
   ADD      r2, r2, 1
   SBCO     r2, CONST_PRURAM, PRU1_BAD_PATTERN_COUNT, 4
noerr:
   LSR      r4, r0, 28           // Get shift count

      // Here we shift WORD1, WORD2 by r4. If we use up all bits in WORD2
      // we read another word into WORD2 and possibly shift it to correct
      // position.
   LSL      WORD1, WORD1, r4
shift:
   MOV      r0, 32
   SUB      r0, r0, r4
   LSR      r2, WORD2, r0                // High bits to move to WORD1
   OR       WORD1, WORD1, r2
   LSL      WORD2, WORD2, r4             // And move next bits to top
   SUB      BIT_COUNT, BIT_COUNT, r4
   QBBC     waitfree, BIT_COUNT, 31      // Branch if bits left in WORD1
   SUB      WORDS_LEFT, WORDS_LEFT, 1
   QBEQ     finished, WORDS_LEFT, 0      // Branch if no words left

      // Get next word and update pointer
   LBBO     WORD2, DDR_ADDR, DDR_OFFSET, 4 
   ADD      DDR_OFFSET, DDR_OFFSET, 4

      // If bitcount is -1 then shift done. Otherwise we need to
      // fix the bits that didn't get shifted into WORD1
   QBEQ     waitfree_setbit_count, BIT_COUNT.b0, 255
      // Number of bits we need to shift into WORD1
   XOR      r4, BIT_COUNT.b0, 255
   MOV      BIT_COUNT, 31                // 32 bits to process
   JMP      shift
waitfree_setbit_count:
   MOV      BIT_COUNT, 31                // 32 bits to process
      // Check to see if space is available before doing next word
waitfree:
      // Get current read location and state of pru 0. If no longer
      // read go handle. Wait for room
   XIN      10, PRU0_BUF_STATE, 4        
   QBNE     check_write, PRU0_STATE, STATE_READ 
   QBNE     bitloop, PRU0_BUF_OFFSET, PRU1_BUF_OFFSET 
      // Indicate buffer has data and PRU 0 can start PWM generation
   MOV      PRU1_STATE, STATE_READ_FILLED
   XOUT     10, PRU1_BUF_STATE, 4                    
   JMP      waitfree

      // Write a zero word to buffer to indicate end of data for track
finished:
   SET      PRU1_BUF_STATE, 16           // Tell PRU 0 we are done writing word
      // We need to set filled to ensure that we don't deadlock waiting for
      // space to write our zero
   MOV      PRU1_STATE, STATE_READ_FILLED
   XOUT     10, PRU1_BUF_STATE, 4
   CLR      PRU1_BUF_STATE, 16
finished2:
   XIN      10, PRU0_BUF_STATE, 4          // Get current read location pru 0
   QBNE     check_write, PRU0_STATE, STATE_READ   // Handle state change
   QBEQ     finished2, PRU0_BUF_OFFSET, PRU1_BUF_OFFSET //  matches wait
   MOV      r2, 0                          // Zero marks end of PWM words
   SBCO     r2, CONST_PRUSHAREDRAM, PRU1_BUF_OFFSET, 4  // Store it
   ADD      PRU1_BUF_OFFSET, PRU1_BUF_OFFSET, 4         // Move to next location
      // Wrap if needed
   AND      PRU1_BUF_OFFSET, PRU1_BUF_OFFSET, SHARED_PWM_READ_MASK
   XOUT     10, PRU1_BUF_STATE, 4
finished2_lp:
   XIN     10, PRU0_BUF_STATE, 4
   QBNE    finished2_lp, PRU0_STATE, STATE_READ_DONE
   JMP      wait_read               // Wait read sets our state to done

      // See if switch to write or terminated was requested. If not restart read.
check_write:
   QBEQ    DONE, PRU0_STATE, STATE_EXIT
      // If PRU 0 stopped read restart, wait_read sets state to done
   QBEQ    wait_read, PRU0_STATE, STATE_READ_DONE
   QBEQ    do_write, PRU0_STATE, STATE_WRITE_WAIT
   HALT
              
do_write:
      // Mark data dirty so ARM can write it to disk.
   MOV     r0, 1
   SBCO    r0, CONST_PRURAM, PRU1_CYL_DIRTY, 4
      // Get bit location write starts from and tell PRU 0 we are ready
   XIN     10, TRACK_BIT, 4         
   MOV     PRU1_STATE, STATE_WRITE_WAIT
   MOV     PRU1_BUF_OFFSET, 0         // Set to start of buffer
   XOUT    10, PRU1_BUF_STATE, 4
      // Wait for PRU 0 to go to write state
write_wait_lp:
   XIN     10, PRU0_BUF_STATE, 4         
   QBNE    write_wait_lp, PRU0_STATE, STATE_WRITE
   MOV     PRU1_STATE, STATE_WRITE
   XOUT    10, PRU1_BUF_STATE, 4

   LSR     DDR_OFFSET, TRACK_BIT, 5 // Convert bit count to word count

   SUB     WORDS_LEFT, TRACK_WORDS, DDR_OFFSET  // Words left
      // Now convert to bit in word. We count from msb bit = 31 - remainder
   LSL     BIT_COUNT, DDR_OFFSET, 5
   SUB     BIT_COUNT, TRACK_BIT, BIT_COUNT      // Remainder
   MOV     r0, 31
   SUB     BIT_COUNT, r0, BIT_COUNT
      // Convert DDR_OFFSET in words to bytes and add to start of track to
      // get offset of words we're writing.
   LSL     DDR_OFFSET, DDR_OFFSET, 2 
   ADD     DDR_OFFSET, DDR_OFFSET, TRACK_OFFSET
      // Clear WORD1 & 2. WORD1 will be a mask of the bits write didn't change
      // WORD2 the changed bits. We need this to combine the modified
      // bits with the unmodified bits in the word at the start and end of
      // the write.
   FILL    &WORD1, 4
   ZERO    &WORD2, 4
delta_loop:
   XIN     10, PRU0_BUF_STATE, 4                
   QBEQ    delta_loop, PRU0_BUF_OFFSET, PRU1_BUF_OFFSET  // Wait for data
      // Get word, stop if zero. Update offset in buffer
   LBCO    r4, CONST_PRUSHAREDRAM, PRU1_BUF_OFFSET, 4
   QBEQ    write_done, r4, 0
   ADD     PRU1_BUF_OFFSET, PRU1_BUF_OFFSET, 4
   AND     PRU1_BUF_OFFSET, PRU1_BUF_OFFSET, SHARED_DELTAS_WRITE_MASK
   XOUT    10, PRU1_BUF_STATE, 4
      // Nominal bit spacing is 20 counts, 100 ns. If time is less than 30
      // we will call it just a one otherwise some zeros precede the one.
   QBGT    done_zeros, r4, 30
      // Write a zero bit, decrement time and check if more zero
wzero:
   CLR     WORD1, WORD1, BIT_COUNT.b0
   SUB     BIT_COUNT, BIT_COUNT, 1
      // If count went negative write the word back to DDR
   QBBC    wnext, BIT_COUNT, 31
   CALL    write_word
wnext:
   SUB     r4, r4, 20
   QBLE    wzero, r4, 30
      // Write a one
done_zeros:
   CLR     WORD1, WORD1, BIT_COUNT.b0
   SET     WORD2, WORD2, BIT_COUNT.b0
   SUB     BIT_COUNT, BIT_COUNT, 1
   QBBC    delta_loop, BIT_COUNT, 31
   CALL    write_word
   JMP     delta_loop

      // Write the word back to DDR. If the mask is all ones we don't
      // need to read the existing word and combine.
write_word:
   QBEQ    just_write, WORD1, 0
   LBBO    r0, DDR_ADDR, DDR_OFFSET, 4     // Get word
   AND     r0, r0, WORD1                   // Clear bits we wish to change
   OR      WORD2, WORD2, r0                // And put them in
just_write:
   SBBO    WORD2, DDR_ADDR, DDR_OFFSET, 4  // Store word
   ADD     DDR_OFFSET, DDR_OFFSET, 4       // Point to next
   SUB     WORDS_LEFT, WORDS_LEFT, 1
   QBNE    write_ret, WORDS_LEFT, 0        // Branch if more words in track
   MOV     DDR_OFFSET, TRACK_OFFSET        // Wrap to beginning of track
   MOV     WORDS_LEFT, TRACK_WORDS
write_ret:
   MOV     BIT_COUNT, 31
   FILL    &WORD1, 4
   ZERO    &WORD2, 4
   RET

write_done:
      // Write word in progress
   CALL    write_word
      // Let PRU 0 continue. 
   MOV      PRU1_STATE, STATE_WRITE_DONE
   XOUT     10, PRU1_BUF_STATE, 4
write_done_wait:
   XIN      10, PRU0_BUF_STATE, 4          // Get pru0 status
   QBNE     write_done_wait, PRU0_STATE, STATE_READ
   JMP      restart_read                   // Restart read from DDR_OFFSET

DONE:
      // Send notification to Host for program completion
   MOV       r31.b0, PRU1_ARM_INTERRUPT+16

      // Halt the processor
   HALT

      // Restarted here if the ARM wishes to abort an operation. Any needed
      // cleanup should be here to restore to a good state. Currently none
   .ORIGIN RESTART_ADDR
      // Send notification to Host for program completion
   MOV      r31.b0, PRU1_ARM_INTERRUPT+16

      // Halt the processor
   HALT

