// This module analyzes the disk format
// Call analyze_disk to perform the analysis
//
// We perform the analysis by trying the various formats we know of until we
// find one that matches and other tests. See the routines for details.

// Copyright 2014 David Gesswein.
// This file is part of MFM disk utilities.
//
// MFM disk utilities is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// MFM disk utilities is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MFM disk utilities.  If not, see <http://www.gnu.org/licenses/>.

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <inttypes.h>

#include "msg.h"
#include "crc_ecc.h"
#include "mfm_decoder.h"
#include "cmd.h"
#include "deltas_read.h"
#include "drive.h"
#include "pru_setup.h"
#include "analyze.h"

// Prints CRC information
static void print_crc_info(CRC_INFO *crc_info, int msg_type) {
   msg(msg_type, "Polynomial %llx length %d initial value %llx\n", crc_info->poly, crc_info->length,
         crc_info->init_value);
}

// Errors we enable from MFM decoding routines while analyzing.
static int decode_errors;

// These are the formats we will search through.
static struct {
uint64_t poly;
int length;
int ecc_span;
} all_poly[] = {
  {0x00a00805, 32, 5},
  {0x1021, 16, 0},
  {0x140a0445, 32, 5},
  {0x140a0444000101, 56, 16}, // From datasheet, not tested
  {0x0104c981, 32, 5}
};
static uint64_t all_init[] = {0, 0xffffffffffffffff, 0x2605fb9c, 0xd4d7ca20};
// TODO: Update to use full enum list
static int all_controller[] =
   {CONTROLLER_WD_1006, CONTROLLER_OMTI_5510, CONTROLLER_DEC_RQDX3,
    CONTROLLER_XEBEC_104786 };
// Smallest sector size should be first in list
static int all_sector_size[] = {256, 512, 1024, 2048, 4096};

// Clear upper bits to make a length bits value
static uint64_t trim_value(uint64_t value, int length) {
   if (length < 64) {
      value &= (((uint64_t) 1 << length) - 1);
   }
   return value;
}


// Print the drive RPM
static void drive_rpm() {
   if (pru_exec_cmd(CMD_RPM, 0) == 0) {
      msg(MSG_INFO, "Drive RPM %.1f\n", 200e6 / pru_get_cmd_data() * 60);
   } else {
      exit(1);
   }
}

// Try to find the controller type (header format) and CRC parameters for the
// header portion of sectors.
// The data to analyze has already been read before routine called.
//
// drive_params: Drive parameters determined so far and return what we have determined
// Cyl and head: What track the data was from.
// deltas: MFM delta time transition data to analyze
// return: 0 ok, 1 multiple matches found, 2 unable to find format
static int analyze_header(DRIVE_PARAMS *drive_params, int cyl, int head, void *deltas)
{
   // Loop variables
   int poly, init, controller;
   int i;
   // The best match CRC and controller type info so far
   CRC_INFO header_crc_info;
   int controller_type = -1;
   // Numbers of good sectors found
   int good_header_count, previous_good_header_count = 0;
   // Variable to restore global error print mask
   int msg_mask_hold;
   // The status of each sector decoded.
   SECTOR_STATUS sector_status_list[MAX_SECTORS];
   // And read status
   SECTOR_DECODE_STATUS status;
   // Return code
   int rc = 0;

   header_crc_info.poly = 0;
   drive_params->num_sectors = MAX_SECTORS;
   drive_params->num_head = MAX_HEAD;
   drive_params->header_crc.ecc_max_span = 0;
   drive_params->data_crc.ecc_max_span = 0;
   // Sector size needs to be smallest value to prevent missing next header.
   // Since we aren't trying to decode data it doesn't matter if it's the wrong value.
   drive_params->sector_size = all_sector_size[0];

   // Try an exhaustive search of all the formats we know about. If we get too
   // many we may have to try something smarter.
   for (poly = 0; poly < ARRAYSIZE(all_poly); poly++) {
      drive_params->header_crc.poly = all_poly[poly].poly;
      drive_params->header_crc.length = all_poly[poly].length;
      drive_params->header_crc.ecc_max_span = all_poly[poly].ecc_span;
      for (init = 0; init < ARRAYSIZE(all_init); init++) {
         drive_params->header_crc.init_value = trim_value(all_init[init],
               drive_params->header_crc.length);
         drive_params->data_crc = drive_params->header_crc;
         for (controller = 0; controller < ARRAYSIZE(all_controller); controller++) {
            // The OMTI_5510 and DEC_RQDX3 have one extra header byte. After
            // the CRC goes to zero when it matches, it will stay zero if another
            // zero byte is added. This will cause the other longer headers to
            // also match even if they aren't proper. If we found the shorter
            // header is good don't check the longer ones. 
            // TODO: Drive this with data instead of hard coding and see
            // if it's really needed now that I changed checking of sector
            // headers
            if (controller_type == CONTROLLER_WD_1006 &&
                  (controller == CONTROLLER_OMTI_5510 || 
                   controller == CONTROLLER_DEC_RQDX3)) {
               break;
            }
            drive_params->controller = all_controller[controller];
            mfm_init_sector_status_list(sector_status_list, drive_params->num_sectors);
            msg_mask_hold = msg_set_err_mask(decode_errors);
            // Decode track
            status = mfm_decode_track(drive_params, cyl, head, deltas, NULL, 
                  sector_status_list);
            msg_set_err_mask(msg_mask_hold);
            if (status & SECT_ZERO_CRC) {
               msg(MSG_DEBUG, "Found zero CRC header controller %s:\n",
                     controller_names[drive_params->controller]);
               print_crc_info(&drive_params->header_crc, MSG_DEBUG);

            }
            // Now find out how many good sectors we got with these parameters
            good_header_count = 0;
            for (i = 0; i < drive_params->num_sectors; i++) {
               if (!(sector_status_list[i].status & SECT_BAD_HEADER)) {
                  good_header_count++;
               }
            }
            // If we found at least 2 sectors
            if (good_header_count >= 2) {
               // If we have a previous match print both
               if (header_crc_info.poly != 0) {
                  msg(MSG_ERR_SERIOUS, "Found multiple matching header parameters. Will use largest matches:\n");
                  msg(MSG_ERR_SERIOUS, "Matches %d controller %s ", good_header_count,
                        controller_names[drive_params->controller]);
                  print_crc_info(&drive_params->header_crc, MSG_ERR_SERIOUS);
                  msg(MSG_ERR_SERIOUS, "Matches %d controller %s ", previous_good_header_count,
                        controller_names[controller_type]);
                  print_crc_info(&header_crc_info, MSG_ERR_SERIOUS);
                  rc = 1;
               }
               // And keep the best
               if (good_header_count > previous_good_header_count) {
                  header_crc_info = drive_params->header_crc;
                  controller_type = drive_params->controller;
                  previous_good_header_count = good_header_count;
               }
            }
         }
      }
   }
   // Print what we found and put it in drive_params
   if (header_crc_info.poly != 0) {
      msg(MSG_INFO, "Header CRC Information:\n");
      print_crc_info(&header_crc_info, MSG_INFO);
      msg(MSG_INFO, "Controller type %s\n", controller_names[controller_type]);
   } else {
      // This is fatal since if we can't continue to next analysis
      msg(MSG_FATAL, "Unable to determine CRC & Controller type\n");
      rc = 2;
   }
   drive_params->header_crc = header_crc_info;
   drive_params->controller = controller_type;
   return rc;
}

// Try to find the sector size and CRC parameters for the data portion of
// sectors.
// The data to analyze has already been read before routine called.
//
// drive_params: Drive parameters determined so far and return what we have determined
// deltas: MFM delta time transition data to analyze
// return: 0 ok, 1 multiple matches found, 2 unable to find format
static int analyze_data(DRIVE_PARAMS *drive_params, int cyl, int head, void *deltas)
{
   // Loop variables
   int poly, init, size_ndx;
   int i;
   // The best match CRC and sector size info so far
   CRC_INFO data_crc_info;
   // And read status
   SECTOR_DECODE_STATUS status;
   int sector_size = 0;
   // Numbers of good sectors found
   int good_data_count, previous_good_data_count = 0;
   // Variable to restore global error print mask
   int msg_mask_hold;
   // The status of each sector decoded.
   SECTOR_STATUS sector_status_list[MAX_SECTORS];
   // Return code
   int rc = 0;

   data_crc_info.poly = 0;
   // Try an exhaustive search of all the formats we know about. If we get too
   // many we may have to try something smarter.
   for (poly = 0; poly < ARRAYSIZE(all_poly); poly++) {
      drive_params->data_crc.poly = all_poly[poly].poly;
      drive_params->data_crc.length = all_poly[poly].length;
      drive_params->data_crc.ecc_max_span = all_poly[poly].ecc_span;
      for (init = 0; init < ARRAYSIZE(all_init); init++) {
         drive_params->data_crc.init_value = trim_value(all_init[init],
               drive_params->data_crc.length);
         for (size_ndx = 0; size_ndx < ARRAYSIZE(all_sector_size); size_ndx++) {
            drive_params->sector_size = all_sector_size[size_ndx];
            mfm_init_sector_status_list(sector_status_list, drive_params->num_sectors);
            msg_mask_hold = msg_set_err_mask(decode_errors);
            // Decode track
            status = mfm_decode_track(drive_params, cyl, head, deltas, NULL, sector_status_list);
            msg_set_err_mask(msg_mask_hold);
            if (status & SECT_ZERO_CRC) {
               msg(MSG_DEBUG, "Found zero CRC data size %d:\n", drive_params->sector_size);
               print_crc_info(&drive_params->data_crc, MSG_DEBUG);

            }
            // Now find out how many good sectors we got with these parameters
            good_data_count = 0;
            for (i = 0; i < drive_params->num_sectors; i++) {
               //printf("status %d %x\n",i,sector_status_list[i].status);
               if (!UNRECOVERED_ERROR(sector_status_list[i].status)) {
                  good_data_count++;
               }
            }
            // If we found a good sector
            if (good_data_count > 0) {
               // If we have a previous match print both
               if (data_crc_info.poly != 0) {
                  msg(MSG_ERR_SERIOUS,
                     "Found multiple matching data CRC parameters. Largest matches will be used:\n");
                  msg(MSG_ERR_SERIOUS, "Matches %d sector size %d ", good_data_count,
                        drive_params->sector_size);
                  print_crc_info(&drive_params->data_crc, MSG_ERR_SERIOUS);
                  msg(MSG_ERR_SERIOUS, "Matches %d sector size %d ", previous_good_data_count,
                        sector_size);
                  print_crc_info(&data_crc_info, MSG_ERR_SERIOUS);
                  rc = 1;
               }
               // And keep the best
               if (good_data_count > previous_good_data_count) {
                  data_crc_info = drive_params->data_crc;
                  sector_size = drive_params->sector_size;
                  previous_good_data_count = good_data_count;
               }
            }
         }
      }
   }
   // Print what we found and put it in drive_params
   if (data_crc_info.poly != 0) {
      msg(MSG_INFO, "Sector length %d, Data CRC Information:\n", sector_size);
      print_crc_info(&data_crc_info, MSG_INFO);
   } else {
      // This is fatal since if we can't continue to next analysis
      msg(MSG_FATAL, "Unable to determine DATA CRC type\n");
      rc = 2;
   }
   drive_params->sector_size = sector_size;
   drive_params->data_crc = data_crc_info;
   return rc;
}

// Determine the number of heads, sectors per track, and interleave pattern
//
// drive_params: Drive parameters determined so far and return what we have determined
// cyl: cylinder the data is from
// deltas: MFM delta time transition data to analyze (filled after read)

// TODO, only handles drives where the interleave pattern is the same on all tracks
// TODO, this won't work when heads >= 8 for WD1003 which truncates head
// number in header to 3 bits.
static void analyze_sectors(DRIVE_PARAMS *drive_params, int cyl, void *deltas) {
   int msg_mask_hold;
   // We return a pointer to this so it must be static (may be better to malloc)
   static uint8_t interleave[MAX_SECTORS];
   int unknown_interleave;
   int head_mismatch = 0;
   SECTOR_STATUS sector_status_list[MAX_SECTORS];
   int head;
   int found_header;
   int unrecovered_error;
   int max_sector, min_sector, last_good_head;
   int i;
   SECTOR_DECODE_STATUS status;

   // TODD, this should also detect the WD controllers that support 16 heads but only
    // put 0-7 in the header.
    max_sector = 0;
    min_sector = MAX_SECTORS;
    last_good_head = -1;
    unrecovered_error = 0;
    unknown_interleave = 0;
    memset(interleave, 255, sizeof(interleave));
    for (head = 0; head < MAX_HEAD && !head_mismatch; head++) {
       int found_bad_header;
       int err_count = 0;
       drive_set_head(head);
       msg_mask_hold = msg_set_err_mask(decode_errors);
       // Try to get a good read. sector_status_list will be the best from
       // all the reads.
       do {
          if (pru_exec_cmd(CMD_READ_TRACK, 0) != 0) {
             drive_print_drive_status(MSG_FATAL, drive_get_drive_status());
             exit(1);
          }
          // OK to start reading deltas
          deltas_start_read(cyl, head);
          mfm_init_sector_status_list(sector_status_list, drive_params->num_sectors);
          status = mfm_decode_track(drive_params, cyl, head, deltas, NULL, sector_status_list);
       } while (UNRECOVERED_ERROR(status) && ++err_count < 16);
       msg_set_err_mask(msg_mask_hold);
       if (UNRECOVERED_ERROR(status)) {
          unrecovered_error = 1;
       }
       found_bad_header = 0;
       found_header = 0;
       for (i = 0; i < drive_params->num_sectors; i++) {
          // If we missed a header after finding one, stop looking at
          // interleave since it is likely to be wrong. With
          // read errors we still may get confused.
          if (sector_status_list[i].status & SECT_BAD_HEADER && found_header) {
             found_bad_header = 1;
          }
          if (sector_status_list[i].status & SECT_HEADER_FOUND) {
             found_header = 1;
             if (!found_bad_header) {
                if (interleave[sector_status_list[i].logical_sector] != 255 &&
                      interleave[sector_status_list[i].logical_sector] !=
                            sector_status_list[i].sector && !unknown_interleave) {
                   msg(MSG_ERR, "Interleave mismatch previous entry %d, %d was %d now %d\n",
                         i, sector_status_list[i].logical_sector,
                         interleave[sector_status_list[i].logical_sector],
                         sector_status_list[i].sector);
                   unknown_interleave = 1;
                }
                interleave[sector_status_list[i].logical_sector] = sector_status_list[i].sector;
             }
             max_sector = MAX(max_sector, i);
             min_sector = MIN(min_sector, sector_status_list[i].sector);
             if (sector_status_list[i].head == head) {
                last_good_head = head;
             } else {
                if (!head_mismatch) {
                   msg(MSG_INFO, "Selected head %d found %d, last good head found %d\n",
                         head, sector_status_list[i].head, last_good_head);
                   head_mismatch = 1;
                }
             }
          }
       }
    }
    // If we had a read error but got some good error warn. If nothing readable
    // assume we were trying to read an invalid head
    if (unrecovered_error && found_header) {
       msg(MSG_ERR, "Read errors trying to determine sector numbering, results may be in error\n");
    }
    if (last_good_head == -1) {
       msg(MSG_FATAL, "Unable to determine number of heads\n");
       exit(1);
    }
    // We store number of heads, not maximum head number (starting at 0)
    drive_params->num_head = last_good_head+1;
    drive_params->num_sectors = max_sector - min_sector + 1;
    drive_params->first_sector_number = min_sector;
    msg(MSG_INFO, "Number of heads %d number of sectors %d first sector %d\n",
          drive_params->num_head, drive_params->num_sectors,
          drive_params->first_sector_number);

    if (unknown_interleave) {
       msg(MSG_ERR, "Unable to determine interleave. Interleave value is not required\n");
       drive_params->sector_numbers = NULL;
    } else {
       msg(MSG_INFO, "Interleave:");
       for (i = 0; i < drive_params->num_sectors; i++) {
          msg(MSG_INFO, " %d",interleave[i]);
       }
       msg(MSG_INFO, "\n");
       drive_params->sector_numbers = interleave;
    }
}

// Determine what seek method should be used for drive
//
// drive_params: Drive parameters determined so far and return what we have determined
// cyl: cylinder the data is from
// deltas: MFM delta time transition data to analyze (filled after read)
static void analyze_seek(DRIVE_PARAMS *drive_params, int start_cyl, int head, void *deltas) {
   int msg_mask_hold;
   SECTOR_STATUS sector_status_list[MAX_SECTORS];
   int i;
   int seek_type;
   int cyl;
   int found_header;

   // We try using the fast seek command to move down the disk and see if
   // we end up on the correct cylinder. If we don't we assume the unbuffered
   // seek method must be used.
   cyl = 30;
   if (pru_exec_cmd(CMD_SEEK_FAST, cyl) != 0) {
      drive_print_drive_status(MSG_FATAL, drive_get_drive_status());
      exit(1);
   }
   if (pru_exec_cmd(CMD_READ_TRACK, 0) != 0) {
      drive_print_drive_status(MSG_FATAL, drive_get_drive_status());
      exit(1);
   }
   // OK to start reading deltas from PRU
   deltas_start_read(start_cyl, head);
   mfm_init_sector_status_list(sector_status_list, drive_params->num_sectors);
   msg_mask_hold = msg_set_err_mask(decode_errors);
   mfm_decode_track(drive_params, start_cyl + cyl, head, deltas, NULL, 
         sector_status_list);
   msg_set_err_mask(msg_mask_hold);
   drive_params->buffered_seek = 1;
   found_header = 0;
   for (i = 0; i < drive_params->num_sectors; i++) {
      if (sector_status_list[i].status & SECT_HEADER_FOUND) {
         found_header = 1;
         if (sector_status_list[i].cyl != cyl + start_cyl) {
            msg(MSG_INFO,"Seek test expected cyl %d got %d\n",cyl, sector_status_list[i].cyl);
            drive_params->buffered_seek = 0;
            cyl = sector_status_list[i].cyl - start_cyl; // Return to start_cyl
         }
      }
   }
   if (!found_header) {
      msg(MSG_FATAL,"No readable data after seek test\n");
      exit(1);
   }
   if (drive_params->buffered_seek) {
      msg(MSG_INFO, "Drive supports buffered seeks (ST412)\n");
      seek_type = CMD_SEEK_FAST;
   } else {
      msg(MSG_INFO, "Drive doesn't support unbuffered seeks (ST506)\n");
      seek_type = CMD_SEEK_SLOW;
   }

   // Return head back to starting cylinder
   if (pru_exec_cmd(seek_type, -cyl) != 0) {
      drive_print_drive_status(MSG_FATAL, drive_get_drive_status());
      exit(1);
   }
}

// Find the number of cylinders the disk has. We step down the disk until we
// find unreadable tracks, the cylinder number in header doesn't match, or
// seek times out. Various disks have different behavior. The tracks past the end
// leading to the park zone are normally not formated. Many drives won't let you
// seek past the last cylinder. They either return to track zero or perform a
// recalibrate which may take enough time for the seek to timeout. Really old
// disk will just hit the end stop so can get the same cylinder. Hitting end stop
// probably isn't great.
//
// drive_params: Drive parameters determined so far and return what we have determined
// start_cyl: cylinder head is on at entry
// head: head currently selected
// deltas: MFM delta time transition data to analyze (filled after read)
//
// TODO, this won't deal well with spared cylinders when they have nonsequential
// values
static void analyze_disk_size(DRIVE_PARAMS *drive_params, int start_cyl, int head, void *deltas) {
   int msg_mask_hold;
   int max_cyl;
   int not_next_cyl, not_next_cyl_count;
   int any_header_found;
   int no_header_count;
   SECTOR_STATUS sector_status_list[MAX_SECTORS];
   int cyl;
   int i;
   int seek_type;
   int last_printed_cyl;

   if (drive_params->buffered_seek) {
      seek_type = CMD_SEEK_FAST;
   } else {
      seek_type = CMD_SEEK_SLOW;
   }

   max_cyl = 0;
   no_header_count = 0;
   not_next_cyl_count = 0;

   for (cyl = start_cyl + 1; cyl < MAX_CYL; cyl++) {
      if (cyl % 5 == 0)
         msg(MSG_PROGRESS, "At cyl %d\r", cyl);

      if (pru_exec_cmd(seek_type, 1) != 0) {
         int wait_count = 0, ready = 0;
         msg(MSG_INFO, "Max cylinder set from drive timeout on seek\n");
         // Wait for seek complete for 5 seconds. This prevents errors if we
         // next read the disk
         while (wait_count++ < 50 && !ready) {
            // Bit is active low
            ready = !(drive_get_drive_status() & BIT_MASK(R31_SEEK_COMPLETE_BIT));
            usleep(100000);
         }
         break;
      }
      if (pru_exec_cmd(CMD_READ_TRACK, 0) != 0) {
         drive_print_drive_status(MSG_FATAL, drive_get_drive_status());
         exit(1);
      }
      // OK to start reading deltas from PRU
      deltas_start_read(cyl, head);

      mfm_init_sector_status_list(sector_status_list, drive_params->num_sectors);
      msg_mask_hold = msg_set_err_mask(decode_errors);
      mfm_decode_track(drive_params, cyl, head, deltas, NULL, sector_status_list);
      msg_set_err_mask(msg_mask_hold);
      any_header_found = 0;
      not_next_cyl = 0;
      last_printed_cyl = -1;
      for (i = 0; i < drive_params->num_sectors; i++) {
         if (sector_status_list[i].status & SECT_HEADER_FOUND) {
            any_header_found = 1;
            max_cyl = MAX(max_cyl, sector_status_list[i].cyl);
            if (cyl != sector_status_list[i].cyl && last_printed_cyl != cyl) {
               msg(MSG_INFO, "Found cylinder %d expected %d\n",sector_status_list[i].cyl, cyl);
               last_printed_cyl = cyl;
               not_next_cyl = 1;
            }
         }
      }
      if (!any_header_found) {
         if (++no_header_count >= 2) {
            msg(MSG_INFO, "Stopping end of disk search due to two unreadable tracks in a row\n");
            break;
         } else {
            msg(MSG_INFO, "No sectors readable from cylinder %d\n",cyl);
         }
      } else {
         no_header_count = 0;
      }
      if (not_next_cyl) {
         if (++not_next_cyl_count >= 2) {
            msg(MSG_INFO, "Stopping end of disk search due to mismatching cylinder count\n");
            break;
         }
      } else {
         not_next_cyl_count = 0;
      }
   }
   drive_params->num_cyl = max_cyl + 1;
   msg(MSG_INFO, "Number of cylinders %d, %.1f MB\n", drive_params->num_cyl,
         (double) drive_params->num_cyl * drive_params->num_head *
         drive_params->num_sectors * drive_params->sector_size / (1000.0*1000));
}

// Routine to determine the various drive parameters. See individual routines
// for details on operation.
//
// drive_params: Return parameters we have determined
// deltas: MFM delta time transition data to analyze (filled after read)
void analyze_disk(DRIVE_PARAMS *drive_params, void *deltas)
{
   int ready = 0;
   int i;
   int rc;
   int cyl = 0, head = 0;
   DRIVE_PARAMS drive_params_hold;

   // Turn off errors from MFM routines while analyzing. Too many
   // messages otherwise.
   decode_errors = MSG_FATAL | (msg_get_err_mask() & MSG_DEBUG);
   // Turn on all errors to see why decoding fails
   //decode_errors = ~1;

   // Save off parameters so we can put back ones we changed as side
   // effects of the analysis.
   drive_params_hold = *drive_params;
   // We don't want to write the analysis information to files
   drive_params->extract_filename = NULL;
   drive_params->transitions_filename = NULL;
   // We will analyze this later
   drive_params->sector_numbers = NULL;
   // Ignore header errors. Was used during testing.
   //drive_params->ignore_header_mismatch = 1;

   // Step through all the head selects to find the drive
   for (i = 1; i <= 4 && !ready; i++) {
      int wait_count = 0;
      drive_params->drive = i;
      drive_select(drive_params->drive);

      // Wait up to 100 ms for drive to be selected
      while (wait_count++ < 100 && !ready) {
         usleep(1000);
         // Bit is active low
         ready = !(drive_get_drive_status() & BIT_MASK(R31_DRIVE_SEL));
      }
   }
   if (ready) {
      msg(MSG_INFO,"Found drive at select %d\n",drive_params->drive);
   } else {
      msg(MSG_INFO,"Unable to find drive. If just powered on retry in 30 seconds\n");
      exit(1);
   }
   drive_setup(drive_params);
   drive_rpm();

   // We don't want to write transitions to file so pass null
   deltas_start_thread(NULL);

   // Don't write decoded data either
   mfm_decode_setup(drive_params, 0);

   // For header and field analysis we read one track and reanalyze the same data
   drive_set_head(head);
   if (pru_exec_cmd(CMD_READ_TRACK, 0) != 0) {
      drive_print_drive_status(MSG_FATAL, drive_get_drive_status());
      exit(1);
   }
   // OK to start reading deltas from PRU
   deltas_start_read(cyl, head);

   rc = analyze_header(drive_params, cyl, head, deltas);
   if (rc != 2) {
      rc |=  analyze_data(drive_params, cyl, head, deltas);
   }
   // If we either got no valid information or multiple possible values try
   // cyl 1 head 1. Cyl 0 head 0 is poor for distinguishing different header
   // formats. Also may help if first track has too many read errors.
   // Analyzing track 0 is useful for detecting the weird multiple formats for
   // the DEC RQDX3 so it is tested first.
   if (rc != 0) {
      msg(MSG_INFO,"Retrying on a different cylinder and head\n");
      if (pru_exec_cmd(CMD_SEEK_SLOW, 1) != 0) {
         drive_print_drive_status(MSG_FATAL, drive_get_drive_status());
         exit(1);
      }
      cyl = 1;
      head = 1;
      drive_set_head(head);
      if (pru_exec_cmd(CMD_READ_TRACK, 0) != 0) {
         drive_print_drive_status(MSG_FATAL, drive_get_drive_status());
         exit(1);
      }
      // OK to start reading deltas from PRU
      deltas_start_read(cyl, head);

      if (analyze_header(drive_params, cyl, head, deltas) == 2) {
         exit(1);
      }
      if (analyze_data(drive_params, cyl, head, deltas) == 2) {
         exit(1);
      }
   }
   analyze_sectors(drive_params, cyl, deltas);
   drive_set_head(head);
   analyze_seek(drive_params, cyl, head, deltas);
   analyze_disk_size(drive_params, cyl, head, deltas);

   // And restore items we changed that aren't determined by the analysis
   drive_params->ignore_header_mismatch = drive_params_hold.ignore_header_mismatch;
   drive_params->extract_filename = drive_params_hold.extract_filename;
   drive_params->transitions_filename = drive_params_hold.transitions_filename;

   deltas_stop_thread();
}
